theory Container_Setup
imports 
  "$AFP/Abstract-Rewriting/SN_Order_Carrier"
  Linear_Polynomial
  Real_Impl
  Linear_Orders
  Forbidden_Patterns
  Tree_Automata
  "$AFP/Containers/Containers" 
begin

(* term, etc *)

derive (eq) ceq "term"
derive (linorder) corder "term"
derive (rbt) set_impl "term"
derive (no) cenum "term"

derive (eq) ceq lab
derive (linorder) corder lab
derive (rbt) set_impl lab
derive (no) cenum lab

derive (eq) ceq ctxt
derive corder ctxt
derive (rbt) set_impl ctxt
derive (no) cenum ctxt

derive (eq) ceq location
derive corder location
derive (dlist) set_impl location
derive (no) cenum location

(* rat *)

derive (eq) ceq rat
derive (no) cenum rat
derive (linorder) corder rat
derive (rbt) set_impl rat

(* real numbers *)

derive (eq) ceq real
derive (linorder) corder real 
derive (rbt) set_impl real
derive (no) cenum real

(* l_poly *)
derive (eq) ceq l_poly
derive corder l_poly
derive (rbt) set_impl l_poly
derive (no) cenum l_poly

(* arctic and arctic_rat *)

derive (eq) ceq arctic
derive corder arctic
derive (rbt) set_impl arctic
derive (no) cenum arctic

derive (eq) ceq arctic_delta
derive corder arctic_delta
derive (rbt) set_impl arctic_delta
derive (no) cenum arctic_delta

(* tree automata *)

lemma infinite_ta_rule_UNIV[simp, intro]: "infinite (UNIV :: ('q,'f)ta_rule set)"
proof -
  fix f :: 'f
  fix q :: 'q
  let ?map = "\<lambda> n. (f (replicate n q) \<rightarrow> q)"
  have "inj ?map" unfolding inj_on_def by auto
  from infinite_super[OF _ range_inj_infinite[OF this]]
  show ?thesis by blast
qed

derive (eq) ceq ta_rule
derive linorder ta_rule
derive (linorder) corder ta_rule
derive (no) cenum ta_rule
derive (rbt) set_impl ta_rule

instantiation ta_rule :: (type,type)finite_UNIV
begin
definition "finite_UNIV = Phantom(('a,'b)ta_rule) False"
instance
  by (intro_classes, unfold finite_UNIV_ta_rule_def, simp)
end

instantiation ta_rule :: (linorder,linorder)cproper_interval
begin
definition "cproper_interval = (\<lambda> ( _ :: ('a,'b)ta_rule option) _ . False)"
instance by (intro_classes, auto)
end

instantiation list :: (type)default
begin
definition "default_list \<equiv> []"
instance ..
end

instantiation lab :: (default,type)default
begin
definition "default_lab = UnLab default_class.default"
instance ..
end

text {* code equation to handle @{term Union} *}
lemma un_foldr[code_unfold]: "\<Union> (set xs) = foldr (op \<union>) xs {}"
  by (induct xs, auto)

text {* do not use @{thm Set_Impl.set_code} equation *}
lemma [code]: "set = foldl (\<lambda> s x. insert x s) {}" (is "_ = foldl ?f _")
proof 
  fix xs
  {
    fix ys
    have "foldl ?f ys xs = set xs \<union> ys"
      by (induct xs arbitrary: ys, auto)
  }
  from this[of "{}"] show "set xs = foldl ?f {} xs" by simp
qed


end
