(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Edg
imports
  Trs
  Ground_Context
begin

no_notation Term.Hole ("\<box>")
notation Ground_Context.GCHole ("\<box>")
notation Ground_Context.equiv_class ("[_]")

(* a semantic version of cap which is not executable *)
fun cap :: "('f, 'v) trs \<Rightarrow> ('f, 'v) term  \<Rightarrow> ('f, 'v) gctxt" where
  "cap _ (Var _)    = \<box>"
| "cap R (Fun f ts) = (
    let h = GCFun f (map (cap R) ts) in
    if (\<exists>r \<in> R. Ground_Context.match h (fst r)) then \<box> else h)"

lemma cap_sound:
  assumes "(s \<cdot> \<sigma>, t) \<in> (rstep R)^*" shows "t \<in> [cap R s]"
using assms
proof (induct s arbitrary: t rule: term_induct)
  case (Var x) thus ?case by simp
next
  case (Fun f ss)
  let ?h = "GCFun f (map (cap R) ss)" 
  show ?case 
  proof (cases "cap R (Fun f ss)")
    case GCHole
    thus ?thesis by simp
  next
    case (GCFun g hs)
    hence res: "cap R (Fun f ss) = ?h"
      by (cases "\<exists> r \<in> R. Ground_Context.match ?h (fst r)", auto simp: Let_def)
    have nrr: "\<And> u. (Fun f ss \<cdot> \<sigma>, u) \<in> (nrrstep R)^* \<Longrightarrow> u \<in> [?h]" 
    proof -
      fix u
      assume "(Fun f ss \<cdot> \<sigma>, u) \<in> (nrrstep R)^*"
      hence cond: "root (Fun f ss \<cdot> \<sigma>) = root u \<and> (\<forall>i<num_args (Fun f ss \<cdot> \<sigma>). (Fun f ss \<cdot> \<sigma> |_ <i>, u |_ <i> ) \<in> (rstep R)^*)"	
	by (rule nrrsteps_imp_eq_root_arg_rsteps)
      hence "root (Fun f ss) = root u" by simp
      hence "\<exists> us. u = Fun f us" by (cases u, auto)
      from this obtain us where u: "u = Fun f us" by force
      with cond have len: "length ss = length us" by simp
      {
	fix i
	assume ilen: "i < length us"
	with len have "<i> \<in> poss (Fun f ss)" by simp
	hence "Fun f ss \<cdot> \<sigma>|_<i> = Fun f ss |_<i> \<cdot> \<sigma>" by simp
	with ilen len cond have one: "(Fun f ss |_<i> \<cdot> \<sigma>, u|_<i>) \<in> (rstep R)^*" by simp	
	from ilen len have "Fun f ss |_<i> \<in> set ss" by simp
	with one Fun have "u|_<i> \<in> [cap R (Fun f ss |_<i>)]" by simp
	with u have "us ! i \<in> [cap R (ss ! i)]" by simp
      }
      with len have "Fun f us \<in> [?h]" 
        by (auto)
      with u show "u \<in> [?h]" by simp
    qed    
    have "rstep R = rrstep R \<union> nrrstep R" by (rule rstep_iff_rrstep_or_nrrstep)
    from this and `(Fun f ss \<cdot> \<sigma>, t) \<in> (rstep R)^*`
    have  "(Fun f ss \<cdot> \<sigma>, t) \<in> (nrrstep R)^* \<or> (Fun f ss \<cdot> \<sigma>, t) \<in> (nrrstep R)^* O (rrstep R) O (rstep R)^*" by (rule firstStep)
    thus ?thesis
    proof 
      assume "(Fun f ss \<cdot> \<sigma>, t) \<in> (nrrstep R)^*"
      with nrr res show ?thesis by (auto)
    next
      assume "(Fun f ss \<cdot> \<sigma>, t) \<in> (nrrstep R)^* O rrstep R O (rstep R)^*"
      from this obtain u v where fu: "(Fun f ss \<cdot> \<sigma>, u) \<in> (nrrstep R)^*" and uv: "(u,v) \<in> rrstep R" by force
      from fu nrr have uh: "u \<in> [?h]" by simp
      from uv have "\<exists> l r \<sigma>. (l,r) \<in> R \<and> l \<cdot> \<sigma> = u \<and> r \<cdot> \<sigma> = v" by (rule rrstep_imp_rule_subst)
      from this obtain l r \<sigma> where lrR: "(l, r) \<in> R" and match: "l \<cdot> \<sigma> = u" by force
      from uh and match  have "Ground_Context.match ?h l" by (unfold match_def, auto)
      with lrR have "cap R (Fun f ss) = \<box>" by (simp add: Let_def, force)
      thus ?thesis by simp
    qed
  qed
qed

lemma cap_instance_subset: "[cap R (t \<cdot> \<sigma>)] \<subseteq> [cap R t]" 
proof (induct t, simp)
  case (Fun f ts)
  let ?h = "GCFun f (map (cap R) ts)"
  let ?hs = "GCFun f (map (cap R) (map (\<lambda> t. (t \<cdot> \<sigma>)) ts))"
  let ?hss = "GCFun f (map (\<lambda> t. cap R (t \<cdot> \<sigma>)) ts)"
  show ?case 
  proof (cases "\<exists> r \<in> R. Ground_Context.match ?h (fst r)", simp)
    case False
    hence nMatch: "\<forall> r \<in> R. \<not> Ground_Context.match ?h (fst r)" by auto
    {
      fix l
      assume "Ground_Context.match ?hs l"
      from this obtain \<delta> where "l \<cdot> \<delta> \<in> [?hs]" unfolding match_def by auto
      hence "l \<cdot> \<delta> \<in> [?h]" using Fun unfolding match_def by force
      hence "Ground_Context.match ?h l" unfolding match_def by auto
    }
    with nMatch have "\<forall> r \<in> R. \<not> Ground_Context.match ?hs (fst r)" by auto
    with `\<not> (\<exists> r \<in> R. Ground_Context.match ?h (fst r))`
    have id: "cap R (Fun f ts) = ?h \<and> cap R (Fun f ts \<cdot> \<sigma>) = ?hss" by auto
    from Fun have "[?hss] \<subseteq> [?h]" by (force simp: nth_mem)
    with id show ?thesis by simp
  qed
qed

lemma cap_refl: "t \<in> [cap R t]"
  by  (rule cap_sound[of _ Var], simp)

lemma cap_lhs:
  assumes "(l, r) \<in> R"
  shows "cap R (l \<cdot> \<sigma>) = \<box>"
proof (cases l)
  case (Fun f ts)
  show ?thesis
  proof (unfold Fun, simp add: Let_def o_def, intro bexI[OF _ assms], simp add: Fun)
    show "Ground_Context.match (GCFun f (map (\<lambda> x. cap R (x \<cdot> \<sigma>)) ts)) (Fun f ts)"
      by  (unfold match_def, rule exI[of _ \<sigma>], unfold equiv_class.simps, simp add: cap_refl)
  qed
next
  case (Var x)
  {
    fix t
    have "cap R t = \<box>"
    proof (induct t)
      case (Var y)
      thus ?case by auto
    next
      case (Fun f ts)
      show ?case
        unfolding cap.simps Let_def
        by (clarsimp, rule bexI[OF _ assms], unfold Var, simp add: match_def,
        rule exI[of _ "(\<lambda>_. Fun f ts)"], simp add: cap_refl)
    qed  
  }
  thus ?thesis by simp
qed

lemma cap_lhs_var:
  assumes left_var: "\<exists>lr \<in> R. is_Var (fst lr)" 
  shows "cap R s = \<box>"
proof -
  from left_var obtain l r where lr: "(l,r) \<in> R" and l: "is_Var l" by auto
  from l obtain x where l: "l = Var x" by (cases l, auto)
  from cap_lhs[OF lr] have "cap R (l \<cdot> (\<lambda> _. s)) = \<box>" by auto
  thus ?thesis unfolding l by auto
qed

lemma match_lhs_var:
  assumes left_var: "\<exists>lr \<in> R. is_Var (fst lr)" 
  shows "Ground_Context.match (cap R s) t"
  unfolding cap_lhs_var[OF left_var] match_def by auto

abbreviation
  cap_below :: "('f, 'v) trs \<Rightarrow> 'f \<Rightarrow> ('f, 'v) term list \<Rightarrow> ('f, 'v) gctxt"
where
  "cap_below R f ts \<equiv> GCFun f (map (cap R) ts)"

fun match_cap_below :: "('f, 'v) term \<Rightarrow> ('f, 'v) trs \<Rightarrow> ('f, 'v) term \<Rightarrow> bool" where
  "match_cap_below l R (Fun f ts) = Ground_Context.match (cap_below R f ts) l"

lemma match_below:
  assumes match: "Ground_Context.match (GCFun f (bef @ s # aft)) l" (is "Ground_Context.match ?s l")
    and subset: "[s] \<subseteq> [t]"
  shows "Ground_Context.match (GCFun f (bef @ t # aft)) l" (is "Ground_Context.match ?t l")
proof -
  from subset have "[?s] \<subseteq> [?t]" by (rule equiv_class_mono)
  with match show ?thesis unfolding match_def by blast
qed

lemma cap_rewrite:
  fixes t :: "('f, 'v) term"
  assumes "(t, s) \<in> rstep R"
  shows "[cap R s] \<subseteq> [cap R t]"
using assms
proof (induct)
  fix C \<sigma> l r assume "(l, r) \<in> R"
  thus "[cap R C\<langle>r \<cdot> \<sigma>\<rangle>] \<subseteq> [cap R C\<langle>l \<cdot> \<sigma>\<rangle>]"
  proof (induct C)
    case (More f bef D aft)
    hence ind: "[cap R D\<langle>r \<cdot> \<sigma>\<rangle>] \<subseteq> [cap R D\<langle>l \<cdot> \<sigma>\<rangle>]" by auto
    let ?h = "\<lambda>t. GCFun f (map (cap R) bef @ cap R D\<langle>t\<rangle> # map (cap R) aft)"
    from ind have res: "\<And>t. Ground_Context.match (?h (r \<cdot> \<sigma>)) t \<Longrightarrow> Ground_Context.match (?h (l \<cdot> \<sigma>)) t"
      using match_below by force
    show ?case
    proof (cases "\<exists>lr\<in>R. Ground_Context.match (?h (l \<cdot> \<sigma>)) (fst lr)")
      case False 
      hence one: "\<not> (\<exists> lr \<in> R. Ground_Context.match (?h (r \<cdot> \<sigma>)) (fst lr))" using res by blast
      from ind have subset: "[?h (r \<cdot> \<sigma>)] \<subseteq> [?h (l \<cdot> \<sigma>)]" by (rule equiv_class_mono)
      with one show ?thesis by (auto simp: Let_def)
    next
      case True thus ?thesis by simp
    qed
  next
    case Hole
    have "cap R (l \<cdot> \<sigma>) = \<box>"
    proof (cases l)
      case (Var x)
      {
        fix h
	from equiv_class_nonempty obtain s :: "('f, 'v) term" where "s \<in> [h]" by auto
	with Var have "l \<cdot> subst x s \<in> [h]" unfolding subst_def by auto
	hence "Ground_Context.match h (fst (l,r))" unfolding match_def by (simp, blast)
      }
      with `(l, r) \<in> R` show ?thesis by (cases "l \<cdot> \<sigma>", auto simp: Let_def, blast)
    next
      case (Fun f ll)
      hence one: "l \<cdot> \<sigma> = Fun f (map (\<lambda> u. u \<cdot> \<sigma>) ll)" (is "_ = Fun f ?ls") by auto
      {
        fix i 
        have "ll ! i \<cdot> \<sigma> \<in> [cap R (ll ! i \<cdot> \<sigma>)]"
          using cap_sound[OF rtrancl_refl, of "ll!i \<cdot> \<sigma>" "Var"] by simp
      }
      with one have "(fst (l,r)) \<cdot> \<sigma> \<in> [GCFun f (map (\<lambda>x. cap R (x \<cdot> \<sigma>)) ll)]" by auto
      with one and `(l, r) \<in> R`
        show ?thesis by (auto simp: Let_def o_def, unfold match_def, blast)
    qed
    thus "[cap R Hole\<langle>r \<cdot> \<sigma>\<rangle>] \<subseteq> [cap R Hole\<langle>l \<cdot> \<sigma>\<rangle>]" by simp
  qed
qed

lemma cap_rewrites:
  assumes  "(s, t) \<in> (rstep R)^*" 
  shows "[cap R t] \<subseteq> [cap R s]"
using assms
proof (induct)
  case base thus ?case by simp
next
  case (step t u)
  from step(3) cap_rewrite[OF step(2)] show ?case by auto
qed

lemma match_cap_sound: 
  assumes derivation: "(t \<cdot> \<delta>, s \<cdot> \<sigma>) \<in> (rstep R)^*"
  shows "Ground_Context.match (cap R t) s"
proof -
  from derivation have "s \<cdot> \<sigma> \<in> [cap R t]" by (rule cap_sound)
  thus ?thesis by (unfold match_def, auto)
qed

lemma match_cap_inv_sound: 
  assumes derivation: "(t \<cdot> \<delta>, s \<cdot> \<sigma>) \<in> (rstep R)^*"
  shows "Ground_Context.match (cap (R^-1) s) t"
proof -
  have id: "((rstep R)^*)^-1 = (rstep (R^-1))^*" unfolding rtrancl_converse[symmetric] reverseTrs ..
  from derivation have "(s \<cdot> \<sigma>, t \<cdot> \<delta>) \<in> ((rstep R)^*)^-1" by auto
  from match_cap_sound[OF this[unfolded id]]
  show "Ground_Context.match (cap (R^-1) s) t" .
qed


(* now that cap has been defined we come to the definition of the 
   estimated dependency graph *)
definition EDG :: "('f, 'v) trs \<Rightarrow> (('f, 'v) rule \<times> ('f, 'v) rule) set" where
  "EDG R \<equiv> {((s, t), (u, v)) | s t u v. Ground_Context.match (cap R t) u \<and> Ground_Context.match (cap (R^-1) u) t}"

lemma EDG:
  assumes seq: "(t \<cdot> \<sigma>, u \<cdot> \<delta>) \<in> (rstep R)^*"
  shows "((s,t),(u,v)) \<in> EDG R"
  unfolding EDG_def
  by (rule, intro exI conjI, rule refl, rule match_cap_sound[OF seq], rule match_cap_inv_sound[OF seq])

no_notation Ground_Context.GCHole ("\<box>")
no_notation Ground_Context.equiv_class ("[_]")
notation Term.Hole ("\<box>")

end
