(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Map_Choice
imports 
  "$AFP/Collections/ICF/impl/RBTMapImpl"
  "$AFP/Collections/ICF/impl/RBTSetImpl"
  Linear_Orders
begin 

class key = linorder

instantiation "term" :: (key, key) key
begin
instance ..
end

instantiation lab :: (key, key) key
begin
instance ..
end

instantiation prod :: (key, key) key
begin
instance ..
end

instantiation list :: (key) key
begin
instance ..
end

instantiation nat :: key
begin
instance ..
end

instantiation int :: key
begin
instance ..
end

instantiation bool :: key
begin
instance ..
end

instantiation char :: key
begin
instance ..
end


definition ceta_map_of :: "('a \<times> 'b) list \<Rightarrow> 'a :: key \<Rightarrow> 'b option" where 
  "ceta_map_of ps = rm.\<alpha> (rm.to_map ps)"

lemma ceta_map_of[simp]: "ceta_map_of ps = map_of ps"
  unfolding ceta_map_of_def rm.correct by simp

fun fun_of_map :: "('a \<Rightarrow> 'b option) \<Rightarrow> 'b \<Rightarrow> ('a \<Rightarrow> 'b)" where 
  "fun_of_map m d a = (case m a of Some b \<Rightarrow> b | None \<Rightarrow> d)"

fun fun_of_map_fun :: "('a \<Rightarrow> 'b option) \<Rightarrow> ('a \<Rightarrow> 'b) \<Rightarrow> ('a \<Rightarrow> 'b)" where 
  "fun_of_map_fun m d a = (case m a of Some b \<Rightarrow> b | None \<Rightarrow> d a)"

fun fun_of_map_fun' :: "('a \<Rightarrow> 'b option) \<Rightarrow> ('a \<Rightarrow> 'c) \<Rightarrow> ('b \<Rightarrow> 'c) \<Rightarrow> ('a \<Rightarrow> 'c)" where 
  "fun_of_map_fun' m d f a = (case m a of Some b \<Rightarrow> f b | None \<Rightarrow> d a)"

definition ceta_set_of :: "('a :: key) list \<Rightarrow> 'a \<Rightarrow> bool" where 
  "ceta_set_of ps \<equiv> let tree = rs.from_list ps in (\<lambda> a. rs.memb a tree)"

lemma ceta_set_of[simp]: "ceta_set_of xs = (\<lambda> y. y \<in> set xs)"
  unfolding ceta_set_of_def Let_def
  by (rule ext, simp add: rs.correct)

definition ceta_list_diff :: "('a :: key) list \<Rightarrow> 'a list \<Rightarrow> 'a list" where 
  "ceta_list_diff xs ys \<equiv> rs.to_list (foldl (\<lambda> a b. rs.delete b a) (rs.from_list xs) ys)"

lemma ceta_list_diff[simp]: "set (ceta_list_diff xs ys) = set xs - set ys" 
proof -
  {
    fix xs
    have "rs.\<alpha> (foldl (\<lambda> a b. rs.delete b a) xs ys) = rs.\<alpha> xs - set ys"
    proof (induct ys arbitrary: xs)
      case (Cons y ys)
      show ?case unfolding foldl.simps o_def
        unfolding Cons by (auto simp: rs.correct)
    qed simp
  } note main = this
  have id: "\<And> xs. set (rs.to_list xs) = rs.\<alpha> xs" by (simp add: rs.correct)
  show ?thesis
    unfolding ceta_list_diff_def
    unfolding id
    unfolding main
    by (simp add: rs.correct)
qed

end

